// Test what MCU is used. with ID
//
// setup self Test for Xiao Nrf52840 BLE & Nrf52840 BLE Sense MCU chips.
// basic test of Digital IO, RGB LED , Buzzer if installed on the Xiao Expansion base check push button and OLED Display & SD card (if installed)
//
//
// Pj Glasso
//
//-----Includes- In Arduino, the #include directive is used to include external libraries or header files in a sketch.
// This gives the programmer access to a large group of standard C libraries and libraries written especially for Arduino.
// use <> for including libraries from sketchs / other libraries, use "" when including a .h
// file in the same directory (such as your own library's .h file from its .cpp file.
// See :https://www.arduino.cc/reference/en/libraries/ for the Libraries available.
//
//
#include <Arduino.h>                  
#include <ArduinoBLE.h>    //Connectivity     Enables Bluetooth® Low Energy connectivity on the Arduino   
#include <LSM6DS3.h>
#include <Wire.h>
//#include <RTClib.h>     //Communication
#include <U8x8lib.h>      //Display
#include <SPI.h>          //Communication
#include <SD.h>           //Memory
#include <PCF8563.h>      //RTC chip on Expansion Brd, with battery backup
//#include <SdFat.h>
//#include <Adafruit_SPIFlash.h>

// use D2 for INT0; attach to CLKOUT pin on RTC
const uint8_t INT_PIN = 1;
// flag to update serial; set in interrupt callback
volatile uint8_t tick_tock = 1;
// INT0 interrupt callback; update tick_tock flag
void set_tick_tock(void) {
  tick_tock = 1;
}
BLEService ledService("19B10000-E8F2-537E-4F6C-D104768A1214"); // Bluetooth® Low Energy LED Service
// Bluetooth® Low Energy LED Switch Characteristic - custom 128-bit UUID, read and writable by central
BLEByteCharacteristic switchCharacteristic("19B10001-E8F2-537E-4F6C-D104768A1214", BLERead | BLEWrite);
//

// set up variables using the SD utility library functions:
Sd2Card card;
SdVolume volume;
SdFile root;
uint32_t volumesize;
const int chipSelect = 2; // SCK : D8, MISO : D9, MOSI : D10, GND : GND, VCC : 3V3
#define SD_CS 2           // SD card chip select
File myFile;
const int ledPin = LED_BUILTIN;     // pin to use for the LED
uint32_t chipId = 0;
int BuzzerPin = A3;                 //A3 , P0.29 PIN 4
int buzzer = BuzzerPin;
word vid;
uint8_t mac_address[6];             // MAC_address ARRAY
char buff[6];
char buffBle[17];
char buffer[80];
char bufferu8[80];

U8X8_SSD1306_128X64_NONAME_HW_I2C u8x8(/* clock=*/ PIN_WIRE_SCL, /* data=*/ PIN_WIRE_SDA, /* reset=*/ U8X8_PIN_NONE);
 // OLEDs without Reset of the Display
PCF8563 pcf;
void setup() {
  Serial.begin(9600);
  delay (2000);
  pinMode(BuzzerPin, OUTPUT);
  Serial.println(); // only for mbed 2.9.x
  Serial.println("Power ON \n ");  // Let's BEGIN!!
  initdisplay();    // OLED Init Display Show "Power ON "
  delay (2000);
  Serial.println("Xiao Expansion board and MCU check & Test compiled on " __DATE__ " at " __TIME__);
  Serial.println("Processor came out of reset.");
  Serial.println();
  pinMode(INT_PIN, INPUT);        // set up interrupt pin
  digitalWrite(INT_PIN, HIGH);    // turn on pullup resistors
  attachInterrupt(digitalPinToInterrupt(INT_PIN), set_tick_tock, RISING); // attach interrupt to set_tick_tock callback on rising edge of INT0

  #if defined(ARDUINO_ARCH_NRF52840) //Is it a Xiao? BLE or Sense
  Serial.println("Selected MCU: nRF52840");
  Serial.println();
    delay(4000);
  u8x8.setCursor(0,3); //POSITION STATUS Word
  u8x8.print("Xiao MCU:");
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(8,5); // Position Data float battery
  u8x8.print("nRF52840");
    delay(2000);
  get_unit_id();
  delay(2000);
  startsound();    //Initial startup Sound & Display Update
  delay(2000);
  getTnow(); 
  delay(2000);
  setupblink();  // RGB LED TEST RED, GREEN, BLUE
  delay(2000);
  u8x8.clearDisplay();
  u8x8.setCursor(0, 0);
  goblu();
  goSd();
  delay(2000);
  showMac(); 
  delay(1000);
  showBle();
 Serial.println("\nAdvertising Bluetooth Device");
  Serial.print("  as: ");
  Serial.println((buffBle));

  #elif defined(ARDUINO_ARCH_ESP32C3)
    Serial.println("Selected MCU: ESP32C3");
    Serial.println();
   for(int i=0; i<17; i=i+8) {
	  chipId |= ((ESP.getEfuseMac() >> (40 - i)) & 0xff) << i;
	  }

	Serial.printf("ESP32 Chip model = %s Rev %d\n", ESP.getChipModel(), ESP.getChipRevision());
	Serial.printf("This chip has %d cores\n", ESP.getChipCores());
  Serial.print("Chip ID: "); Serial.println(chipId);

  #elif defined(ARDUINO_ARCH_ESP32S3)
    Serial.println("Selected MCU: ESP32S3");
     Serial.println();
   for(int i=0; i<17; i=i+8) {
	  chipId |= ((ESP.getEfuseMac() >> (40 - i)) & 0xff) << i;
	}

	Serial.printf("ESP32 Chip model = %s Rev %d\n", ESP.getChipModel(), ESP.getChipRevision());
	Serial.printf("This chip has %d cores\n", ESP.getChipCores());
  Serial.print("Chip ID: "); Serial.println(chipId);

  #else
    Serial.println("Selected MCU: Unknown");
  #endif
}

void loop() {
   // check if time display should be output
  if(tick_tock) {
       tick_tock = 0;
  }
  getTonly();
  getBattery();
  showBle();
  showSd();

 //  
 // Your main code here
 //
  BLEDevice central = BLE.central();            // listen for Bluetooth® Low Energy peripherals to connect:
   if (central) {                                // if a central is connected to peripheral:
    Serial.print("Connected to central: ");
     u8x8.setCursor(0,3); //POSITION STATUS Word
     u8x8.print("Xiao BLE:");
     u8x8.clearLine(5);
     u8x8.clearLine(6);
     u8x8.setCursor(4,5); // Position Data float battery
     u8x8.print("Connected");
    // print the central's MAC address:
    Serial.println(central.address());
    digitalWrite(LEDB, LOW);     // turn ON blue LED
    digitalWrite(LEDG, HIGH);   // while the central is still connected to peripheral:
    while (central.connected()) {
        if(tick_tock) {
         tick_tock = 0;
         }
       // getTonly();
      // if the remote device wrote to the characteristic,
      // use the value to control the LED:
      if (switchCharacteristic.written()) {
        if (switchCharacteristic.value()) {   // any value other than 0
          Serial.println(F("LED on"));
          digitalWrite(LEDB, HIGH);
          digitalWrite(LEDG, LOW);         // will turn the LED on.. changed from HIGH to LOW   
        } else {                              // a 0 value
          Serial.println(F("LED off"));
          digitalWrite(LEDG, HIGH);
          digitalWrite(LEDB, LOW);     // turn ON blue LED BLE is ON
        }
          
      }
       //getBattery();
       //showBle();
       //showSd();
      
     }

   // when the central disconnects, print it out:
   Serial.print(F("Disconnected from central: "));
   Serial.println(central.address());
   u8x8.setCursor(0,3); //POSITION STATUS Word
   u8x8.print("Xiao BLE:");
   u8x8.clearLine(5);
   u8x8.clearLine(6);
   u8x8.setCursor(4,5); // Position Data float battery
   u8x8.print("Disconnected");
   digitalWrite(LEDB, HIGH);
   digitalWrite(LEDG, HIGH);
  }
}
 //END

//
// -------------------- Utilities ------------------------- //
//
void testFlash(){

}

void goSd(){
    u8x8.setCursor(0,3); //POSITION STATUS Word
    u8x8.print("Xiao SD:");
    u8x8.clearLine(5);
    u8x8.clearLine(6);
    u8x8.setCursor(4,5); // Position Data float battery
    u8x8.print("TESTING ");
    Serial.print("\nInitializing SD card...");         //initialization of SD and test file open
    if(!SD.begin(32000000UL, SD_CS)) {                   // default 4MHz    
    Serial.println("SD card initialization error");
       return;                                        // [3] SD card initialization error 
     }
  
  Serial.println("Card Mounted");
   myFile = SD.open("test.txt", FILE_WRITE);
  myFile.close();
  u8x8.setCursor(14, 3);
  u8x8.print("SD" );
  //Serial.print("Total space: %lluMB\n", SD.totalBytes() / (1024 * 1024));
  delay(250);
  //Serial.print("Used space: %lluMB\n", SD.usedBytes() / (1024 * 1024));
  
  //Check to see if the file exists:
  if (SD.exists("test.txt"))  //Serial.println("example.txt exists.");
   {
    tone (buzzer, 800);
    delay (220);
    noTone(buzzer);
    delay (20);
    tone (buzzer, 990);
    delay (420);
    noTone(buzzer);
    delay (20);
   } 
  else 
   {
  Serial.println("test.txt doesn't exist.");
  delay(2000);
  myFile = SD.open("test.txt", FILE_WRITE);
  for(byte idx = 0; idx < 20; idx ++)
  {
    myFile.print("abc ");
    myFile.println(idx);
  }
  myFile.close();
  Serial.println("DONE");
  }
  // we'll use the initialization code from the utility libraries
  // since we're just testing if the card is working!
  // if (!card.init(SPI_HALF_SPEED, chipSelect)) {    
  if (!card.init(SPI_FULL_SPEED, chipSelect)) {
   u8x8.setCursor(4,5); // Position Data float battery
   u8x8.print("NO CARD!! ");
   //while (1);
  } else {
    //Serial.println("Wiring is correct and a card is present.");
    u8x8.setCursor(3,5); // Position Data float battery
    u8x8.print("Card Found!! ");
    delay(1000);
  }
  // print the type of card
  //Serial.println();
  Serial.print("Card type: ");
  switch (card.type()) {
    case SD_CARD_TYPE_SD1:
      Serial.println("SD1");
      break;
    case SD_CARD_TYPE_SD2:
      Serial.println("SD2");
      break;
    case SD_CARD_TYPE_SDHC:
      Serial.println("SDHC");
      break;
    default:
      Serial.println("Unknown");
   }
    if (!volume.init(card)) {                 // Now we will try to open the 'volume'/'partition' - it should be FAT16 or FAT32
    Serial.println("Could not find FAT16/FAT32 partition.\nMake sure you've formatted the card");
    while (1);
   }
  // uint32_t volumesize;
  Serial.print("Volume type is:    FAT");
  Serial.println(volume.fatType(), DEC);
  volumesize = volume.blocksPerCluster();    // clusters are collections of blocks
  volumesize *= volume.clusterCount();       // we'll have a lot of clusters
  volumesize /= 2;                           // SD card blocks are always 512 bytes (2 blocks are 1KB)
  volumesize /= 1024;
  Serial.print("SD Volume size (Gb):  ");
  Serial.println((float)volumesize / 1024.0);  // prints Size of SD card
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(3,5); // Position Data float battery
  u8x8.print((float)volumesize / 1024.0);
  u8x8.print("(Gb)");
  delay(2000);
}

void showSd(){
  if (!volume.init(card)) {                 // Now we will try to open the 'volume'/'partition' - it should be FAT16 or FAT32
   // Serial.println("Could not find FAT16/FAT32 partition.\nMake sure you've formatted the card");
    u8x8.setCursor(0,3); //POSITION STATUS Word
    u8x8.print("Xiao SD :");
    u8x8.clearLine(5);
    u8x8.clearLine(6);
    u8x8.setCursor(4,5); // Position Data float battery
    u8x8.print("No Card");
    delay(500);
    return;
  }
  u8x8.setCursor(0,3); //POSITION STATUS Word
  u8x8.print("Xiao SD :");
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(4,5); // Position Data float battery
  u8x8.print((float)volumesize / 1024.0);
  u8x8.print("(Gb)");
  delay(500);
}

void goblu(){
// begin BLE initialization
  if (!BLE.begin()){
    Serial.println("starting Bluetooth® Low Energy module failed!");
    while (1);
  }
  get_id_address(mac_address);
  sprintf(buffBle, "Xiao %02lX%02lX",mac_address[0],mac_address[1]);
    // set advertised local name and service UUID:
  BLE.setLocalName(buffBle);
  BLE.setAdvertisedService(ledService);
  // add the characteristic to the service
  ledService.addCharacteristic(switchCharacteristic);
  // add service
  BLE.addService(ledService);
  // set the initial value for the characeristic:
  switchCharacteristic.writeValue(0);
  BLE.advertise();
  //get_id_address(mac_address);
  //sprintf(buffBle, "Xiao %02lX%02lX",mac_address[0],mac_address[1]);//Serial.println("BLE Peripheral");
}

void getTnow() {
  u8x8.clearDisplay();
  u8x8.setCursor(0, 0);
  Time nowTime = pcf.getTime();//get current time
  u8x8.setCursor(0, 0);
  u8x8.print(nowTime.month);
  u8x8.print("/");
  u8x8.print(nowTime.day);
  u8x8.print("/");
  u8x8.print("20");
  u8x8.print(nowTime.year);
  u8x8.setCursor(0, 3);
  u8x8.print(nowTime.hour );
  u8x8.print(":");
  u8x8.print(nowTime.minute);
  u8x8.print(":");
  u8x8.println(nowTime.second);
  delay(250);
  tick_tock = 1;
  Serial.print("RTC Date & Time is:");          //print current time
  Serial.print(nowTime.month);
  Serial.print("/");
  Serial.print(nowTime.day);
  Serial.print("/");
  Serial.print(nowTime.year);
  Serial.print(" ");
  Serial.print(nowTime.hour);
  Serial.print(":");
  Serial.print(nowTime.minute);
  Serial.print(":");
  Serial.println(nowTime.second);
  delay(250);
  
}

void getTonly() {
  u8x8.setCursor(0, 0);
  u8x8.print(" RTC:");
  Time nowTime = pcf.getTime();//get current time
  u8x8.setCursor(8, 0);
  u8x8.print(nowTime.hour );
  u8x8.print(":");
  u8x8.print(nowTime.minute);
  u8x8.print(":");
  u8x8.println(nowTime.second);
  delay(250);
  tick_tock = 1;
}

void showBle(){
  u8x8.setFont(u8x8_font_8x13B_1x2_r);
  u8x8.setCursor(0,3);
  u8x8.setCursor(0,3); //POSITION STATUS Word
  u8x8.print("Xiao BLE:");
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(0,5); // Position
  u8x8.print("BLE Peripheral");
delay(250);
}

void getBattery(){
  //int Vadc = 3.85;
  float Vbatt = 3.85;// ((510e3 + 1000e3) / 510e3) * 2.4 * Vadc / 4096;
  //Serial.print("Battery Level "); 
  //Serial.println( Vbatt, 3); 
  digitalWrite(LED_GREEN, LOW);
  //u8x8.clearDisplay();
  u8x8.setCursor(0,3); //POSITION STATUS Word
  u8x8.print("Xiao BAT:");
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(9,5); // Position Data float battery
  u8x8.print(Vbatt, 2);
  digitalWrite(LED_GREEN, HIGH);
delay(250);
}

void setupblink(){
  Serial.print("Testing R G B LED - ");
   u8x8.setCursor(0,3); //POSITION STATUS Word
   u8x8.print("Testing LED ");
   u8x8.clearLine(5);
   u8x8.clearLine(6);
   u8x8.setCursor(5,5); // Position Data float battery
   u8x8.print("R.G.B.LED");
   delay(2000);
   setLedRGB(true, false, false);  // Red
   Serial.print("RED, ");
   u8x8.clearLine(5);
   u8x8.clearLine(6);
   u8x8.setCursor(2,5); // Position Data float battery
   u8x8.print("R E D" );
   delay(2000);
   setLedRGB(false, true, false);  // Gren
   Serial.print("GREEN, ");
   u8x8.setCursor(2,5); // Position Data float battery
   u8x8.print("G R E E N");
   delay(2000);
   u8x8.clearLine(5);
   u8x8.clearLine(6);
   setLedRGB(false, false, true);  // Blue 
   Serial.println("BLUE... "); 
   u8x8.setCursor(2,5); // Position Data float battery
   u8x8.print("B L U E ");
   delay(2000);
   setLedRGB(false, false, false);  // OFF
}



void get_unit_id(){
 get_id_address(mac_address);
 sprintf(buff, "Xiao Unit: %02lX%02lX\n",mac_address[0],mac_address[1]);
 Serial.print(buff);
 u8x8.clearDisplay();
 u8x8.setCursor(0,0);
 u8x8.print(buff);
 u8x8.setCursor(6,6);
 delay(2000);
 }

void do_mac(){
  char buffer[80];
  char bufferu8[80];
  Serial.println("");
  get_mac_address(mac_address);
  sprintf(buffer, "Xiao MAC: %02lX:%02lX:%02lX:%02lX:%02lX:%02lX\n",
          mac_address[0],mac_address[1],mac_address[2],mac_address[3],mac_address[4],mac_address[5] ,mac_address[6]);
  Serial.print(buffer);
  sprintf(bufferu8, "Xiao MAC: \n%02lX%02lX%02lX%02lX%02lX%02lX\n",
          mac_address[0],mac_address[1],mac_address[2],mac_address[3],mac_address[4],mac_address[5] ,mac_address[6]);
  u8x8.setCursor(0,3);
  u8x8.print(bufferu8);
}

void get_id_address(uint8_t mac_address[6]) {
	unsigned int device_addr_0 = NRF_FICR->DEVICEADDR[0];
	unsigned int device_addr_1 = NRF_FICR->DEVICEADDR[1];
	const uint8_t* part_0 = reinterpret_cast<const uint8_t*>(&device_addr_0);
	const uint8_t* part_1 = reinterpret_cast<const uint8_t*>(&device_addr_1);
	mac_address[0] = part_1[3];   //changed from get_mac_address elements 0,1
	mac_address[1] = part_1[2];   //
}

void showMac(){
  get_mac_address(mac_address);
 sprintf(buffer, "\nXiao MAC: %02lX:%02lX:%02lX:%02lX:%02lX:%02lX\n", 
          mac_address[0],mac_address[1],mac_address[2],mac_address[3],mac_address[4],mac_address[5] ,mac_address[6]);
  Serial.print(buffer);
 sprintf(bufferu8, "Xiao MAC: \n%02lX%02lX%02lX%02lX%02lX%02lX",
          mac_address[0],mac_address[1],mac_address[2],mac_address[3],mac_address[4],mac_address[5] ,mac_address[6]);
  u8x8.setCursor(0,3);
  u8x8.clearLine(3);
  u8x8.clearLine(4);
  u8x8.clearLine(5);
  u8x8.clearLine(6);   
  u8x8.print(bufferu8);
}

void get_mac_address(uint8_t mac_address[6]) {
	unsigned int device_addr_0 = NRF_FICR->DEVICEADDR[0];
	unsigned int device_addr_1 = NRF_FICR->DEVICEADDR[1];
	const uint8_t* part_0 = reinterpret_cast<const uint8_t*>(&device_addr_0);
	const uint8_t* part_1 = reinterpret_cast<const uint8_t*>(&device_addr_1);
	mac_address[0] = part_1[1];
	mac_address[1] = part_1[0];
	mac_address[2] = part_0[3];
	mac_address[3] = part_0[2];
	mac_address[4] = part_0[1];
	mac_address[5] = part_0[0];
}

void initdisplay() {
  pinMode(LEDR, OUTPUT);  // initialize the LED pin as an output:
  pinMode(LEDG, OUTPUT);  // initialize the LED pin as an output:
  pinMode(LEDB, OUTPUT);
  pinMode(LED_BUILTIN, OUTPUT);  // initialize the LED pin as an output:
  digitalWrite(LEDR, HIGH);
  digitalWrite(LEDG, HIGH);
  digitalWrite(LEDB, HIGH);
  u8x8.begin();
  u8x8.setFlipMode(0);  // set number from 1 to 3, the screen word will rotary 180
  u8x8.setFont(u8x8_font_8x13B_1x2_r);
  u8x8.clearDisplay();
  u8x8.setCursor(0, 0);
  u8x8.print("Power ON ");
}

void startsound() {
  u8x8.setCursor(0,3); //POSITION STATUS Word
  u8x8.print("ROLL Sound:");
  u8x8.clearLine(5);
  u8x8.clearLine(6);
  u8x8.setCursor(8,5); // Position Data float battery
  delay(500);
  u8x8.print("Start-up");
  Serial.println("Playing startup sound");

  tone (buzzer, 890);
    delay (220);
    noTone(buzzer);
    delay (20);
    tone (buzzer, 800);
    delay (220);
    noTone(buzzer);
    delay (20);
    tone (buzzer, 800);
    delay (220);
    noTone(buzzer);
    delay (20);
    tone (buzzer, 990);
    delay (420);
    noTone(buzzer);
    delay (20);
}

void setLedRGB(bool red, bool green, bool blue) {
   if (!red) { digitalWrite(LEDR, HIGH); } else { digitalWrite(LEDR, LOW); }
   if (!green) { digitalWrite(LEDG, HIGH); } else { digitalWrite(LEDG, LOW); }
    if (!blue) { digitalWrite(LEDB, HIGH); } else { digitalWrite(LEDB, LOW); }
}